#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

void send_string(Display *dpy, Window win, const char *str);
void send_char(Display *dpy, Window win, char c);
Window win_from_name(Display *dpy, Window par, const char *name);

int main(int argc, char **argv)
{
	int i;
	Display *dpy;
	Window win = 0;
	char *name = 0;

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 'w':
				win = strtol(argv[++i], 0, 16);
				break;

			case 'n':
				name = argv[++i];
				break;

			default:
				fprintf(stderr, "unrecognized argument: %s\n", argv[i]);
				return 1;
			}
		} else {
			fprintf(stderr, "unexpected argument: %s\n", argv[i]);
			return 1;
		}
	}

	if(!(dpy = XOpenDisplay(0))) {
		fprintf(stderr, "failed to open X11 display\n");
		return 1;
	}

	if(name) {
		if(!(win = win_from_name(dpy, DefaultRootWindow(dpy), name))) {
			fprintf(stderr, "failed to find a window with WM name: %s\n", name);
			XCloseDisplay(dpy);
			return 1;
		}
	}

	if(!win) {
		fprintf(stderr, "You must specify a window to send the events\n");
		XCloseDisplay(dpy);
		return 1;
	}

	send_string(dpy, win, "hello world!\n");

	XCloseDisplay(dpy);
	return 0;
}

void send_string(Display *dpy, Window win, const char *str)
{
	while(*str) {
		send_char(dpy, win, *str++);
	}
}

void send_char(Display *dpy, Window win, char c)
{
	XEvent xev;

	memset(&xev, 0, sizeof xev);
	xev.type = KeyPress;
	xev.xkey.window = win;
	xev.xkey.time = CurrentTime;
	xev.xkey.state = 0;
	xev.xkey.keycode = XKeysymToKeycode(dpy, c);

	XSendEvent(dpy, win, False, KeyPressMask, &xev);

	xev.type = KeyRelease;
	XSendEvent(dpy, win, False, KeyReleaseMask, &xev);
}

Window win_from_name(Display *dpy, Window pwin, const char *name)
{
	int i, nchld;
	Window win, root, par;
	Window *chld;
	XTextProperty prop;

	if(!XQueryTree(dpy, pwin, &root, &par, &chld, &nchld)) {
		fprintf(stderr, "XQueryTree failed\n");
		return 0;
	}

	if(!chld) {
		return 0;
	}

	for(i=0; i<nchld; i++) {
		if(XGetWMName(dpy, chld[i], &prop)) {
			printf("got name: %s\n", prop.value);
			if(strcmp(name, prop.value) == 0) {
				return chld[i];
			}
		}
	}

	for(i=0; i<nchld; i++) {
		if((win = win_from_name(dpy, chld[i], name))) {
			XFree(chld);
			return win;
		}
	}
	XFree(chld);
	return 0;
}
